'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.ocel.link_analysis.variants import classic
from enum import Enum
from pm4py.util import exec_utils
import pandas as pd
from typing import Optional, Dict, Any
from pm4py.objects.log.obj import EventLog, EventStream
from typing import Union
from pm4py.objects.conversion.log import converter


class Variants(Enum):
    CLASSIC = classic


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pd.DataFrame:
    """
    Applies a link analysis algorithm on the provided log object.

    Parameters
    -----------------
    log
        Event log
    variant
        Variant of the algorithm to consider
    parameters
        Variant-specific parameters

    Returns
    -----------------
    link_analysis_dataframe
        Link analysis dataframe
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(
        converter.apply(
            log,
            variant=converter.Variants.TO_DATA_FRAME,
            parameters=parameters,
        ),
        parameters=parameters,
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import shutil
import tempfile
from copy import copy
from enum import Enum

import matplotlib

from pm4py.util import exec_utils, vis_utils
from pm4py.objects.org.sna.obj import SNA
from pm4py.util import constants
from pm4py.util import nx_utils
import networkx as nx


class Parameters(Enum):
    WEIGHT_THRESHOLD = "weight_threshold"
    FORMAT = "format"


def get_temp_file_name(format):
    """
    Gets a temporary file name for the image

    Parameters
    ------------
    format
        Format of the target image
    """
    filename = tempfile.NamedTemporaryFile(suffix="." + format)
    filename.close()

    return filename.name


def apply(sna: SNA, parameters=None):
    """
    Perform SNA visualization starting from the Matrix Container object
    and the Resource-Resource matrix

    Parameters
    -------------
    sna
        Value of the metrics
    parameters
        Possible parameters of the algorithm, including:
            - Parameters.WEIGHT_THRESHOLD -> the weight threshold to use in displaying the graph
            - Parameters.FORMAT -> format of the output image (png, svg ...)

    Returns
    -------------
    temp_file_name
        Name of a temporary file where the visualization is placed
    """
    if parameters is None:
        parameters = {}

    weight_threshold = exec_utils.get_param_value(
        Parameters.WEIGHT_THRESHOLD, parameters, 0
    )
    format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")

    directed = sna.is_directed

    temp_file_name = get_temp_file_name(format)

    if directed:
        graph = nx_utils.DiGraph()
    else:
        graph = nx_utils.Graph()

    connections = {
        x for x, y in sna.connections.items() if y >= weight_threshold
    }

    graph.add_edges_from(connections)

    current_backend = copy(matplotlib.get_backend())
    matplotlib.use("Agg")
    from matplotlib import pyplot

    pyplot.clf()
    nx.draw(
        graph, node_size=500, with_labels=True, pos=nx.circular_layout(graph)
    )
    pyplot.savefig(temp_file_name, bbox_inches="tight")
    pyplot.clf()

    matplotlib.use(current_backend)

    return temp_file_name


def view(temp_file_name, parameters=None):
    """
    View the SNA visualization on the screen

    Parameters
    -------------
    temp_file_name
        Temporary file name
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        if constants.DEFAULT_GVIZ_VIEW == "matplotlib_view":
            import matplotlib.pyplot as plt
            import matplotlib.image as mpimg

            img = mpimg.imread(temp_file_name)
            plt.axis("off")
            plt.tight_layout(pad=0, w_pad=0, h_pad=0)
            plt.imshow(img)
            plt.show()
            return

        if vis_utils.check_visualization_inside_jupyter():
            vis_utils.view_image_in_jupyter(temp_file_name)
        else:
            vis_utils.open_opsystem_image_viewer(temp_file_name)


def save(temp_file_name, dest_file, parameters=None):
    """
    Save the SNA visualization from a temporary file to a well-defined destination file

    Parameters
    -------------
    temp_file_name
        Temporary file name
    dest_file
        Destination file
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    shutil.copyfile(temp_file_name, dest_file)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from typing import Tuple, List, Optional, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureUVCL,
    IMDataStructureDFG,
)
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough
from pm4py.objects.process_tree.obj import ProcessTree, Operator
from pm4py.objects.dfg.obj import DFG
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from copy import copy


class EmptyTracesUVCL(FallThrough[IMDataStructureUVCL]):

    @classmethod
    def apply(
            cls,
            obj: IMDataStructureUVCL,
            pool=None,
            manager=None,
            parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[ProcessTree, List[IMDataStructureUVCL]]]:
        if cls.holds(obj, parameters):
            data_structure = copy(obj.data_structure)
            del data_structure[()]
            if data_structure:
                return ProcessTree(operator=Operator.XOR), [
                    IMDataStructureUVCL(Counter()),
                    IMDataStructureUVCL(data_structure),
                ]
            else:
                return ProcessTree(), []
        else:
            return None

    @classmethod
    def holds(
            cls,
            obj: IMDataStructureUVCL,
            parameters: Optional[Dict[str, Any]] = None,
    ) -> bool:
        return len(list(filter(lambda t: len(t) == 0, obj.data_structure))) > 0


class EmptyTracesDFG(FallThrough[IMDataStructureDFG]):
    @classmethod
    def apply(
            cls,
            obj: IMDataStructureDFG,
            pool=None,
            manager=None,
            parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[ProcessTree, List[IMDataStructureDFG]]]:
        if cls.holds(obj, parameters):
            # If the DFG itself is empty, the log contains only empty traces -> just τ
            dfg = obj.data_structure.dfg
            is_empty = (
                    len(dfg.graph) == 0
                    and len(dfg.start_activities) == 0
                    and len(dfg.end_activities) == 0
            )

            if is_empty:
                return ProcessTree(), []
                # Otherwise, split: XOR(τ, non-empty part)

            return ProcessTree(operator=Operator.XOR), [
                IMDataStructureDFG(InductiveDFG(DFG())),
                # note: skip intentionally False here; τ is already handled on the XOR
                IMDataStructureDFG(InductiveDFG(obj.data_structure.dfg)),
            ]
        return None

    @classmethod
    def holds(
            cls,
            obj: IMDataStructureDFG,
            parameters: Optional[Dict[str, Any]] = None,
    ) -> bool:
        return obj.data_structure.skip



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import tempfile

from graphviz import Digraph

from pm4py.objects.petri_net.obj import Marking, PetriNet
from pm4py.objects.petri_net import properties as petri_properties
from pm4py.util import exec_utils, constants
from enum import Enum
from typing import List, Tuple, Dict
from collections import defaultdict, deque
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ACTIVITY_KEY,
    PARAMETER_CONSTANT_TIMESTAMP_KEY,
)


class Parameters(Enum):
    FORMAT = "format"
    DEBUG = "debug"
    RANKDIR = "set_rankdir"
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    AGGREGATION_MEASURE = "aggregationMeasure"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"
    DECORATIONS = "decorations"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def sort_petri_net(
    transitions: List[PetriNet.Transition],
    places: List[PetriNet.Place],
    arcs: List[PetriNet.Arc],
    initial_marking: Dict[PetriNet.Place, int],
    final_marking: Dict[PetriNet.Place, int],
) -> Tuple[
    List[PetriNet.Transition], List[PetriNet.Place], List[PetriNet.Arc]
]:
    """
    Sorts the Petri net elements based on reachability from/to the initial/final marking

    Parameters
    ----------------
    transitions
        List of Petri net transitions
    places
        List of Petri net places
    arcs
        List of Petri net arcs
    initial_marking
        Initial marking
    final_marking
        Final marking

    Returns
    -----------------
    sorted_transitions
        Sorted list of Petri net transitions
    sorted_places
        Sorted list of Petri net places
    sorted_arcs
        Sorted list of Petri net arcs
    """
    # create adjacency lists for places and transitions
    place_to_transition = defaultdict(list)
    transition_to_place = defaultdict(list)

    for arc in arcs:
        if arc.source in places and arc.target in transitions:
            place_to_transition[arc.source].append(arc.target)
        else:
            transition_to_place[arc.source].append(arc.target)

    # initialize distance dictionaries
    place_distance = {place: float("inf") for place in places}
    transition_distance = {
        transition: float("inf") for transition in transitions
    }

    # initialize the queue with initial marking places
    queue = deque([place for place in initial_marking])

    # set the initial distances
    for place in initial_marking:
        place_distance[place] = 0

    # perform BFS to calculate distances from the initial marking
    while queue:
        current = queue.popleft()
        current_distance = place_distance[current]

        for transition in place_to_transition[current]:
            if transition_distance[transition] > current_distance + 1:
                transition_distance[transition] = current_distance + 1
                for place in transition_to_place[transition]:
                    if (
                        place_distance[place]
                        > transition_distance[transition] + 1
                    ):
                        place_distance[place] = (
                            transition_distance[transition] + 1
                        )
                        queue.append(place)

    # calculate distance from the final marking for sorting purposes
    # initialize distance dictionaries
    place_distance_final = {place: float("inf") for place in places}
    transition_distance_final = {
        transition: float("inf") for transition in transitions
    }

    # initialize the queue with final marking places
    queue = deque([place for place in final_marking])

    # set the initial distances
    for place in final_marking:
        place_distance_final[place] = 0

    # perform BFS to calculate distances from the final marking
    while queue:
        current = queue.popleft()
        current_distance = place_distance_final[current]

        for transition in transition_to_place[current]:
            if transition_distance_final[transition] > current_distance + 1:
                transition_distance_final[transition] = current_distance + 1
                for place in place_to_transition[transition]:
                    if (
                        place_distance_final[place]
                        > transition_distance_final[transition] + 1
                    ):
                        place_distance_final[place] = (
                            transition_distance_final[transition] + 1
                        )
                        queue.append(place)

    # sort places, transitions, and arcs based on distances
    def get_place_priority(place):
        return (place_distance[place], -place_distance_final[place])

    def get_transition_priority(transition):
        return (
            transition_distance[transition],
            -transition_distance_final[transition],
        )

    sorted_places = sorted(places, key=get_place_priority)
    sorted_transitions = sorted(transitions, key=get_transition_priority)
    sorted_arcs = sorted(
        arcs,
        key=lambda x: (
            (
                get_place_priority(x.source)
                if x.source in places
                else get_transition_priority(x.source)
            ),
            (
                get_place_priority(x.target)
                if x.target in places
                else get_transition_priority(x.target)
            ),
        ),
    )

    return sorted_transitions, sorted_places, sorted_arcs


def apply(
    net, initial_marking, final_marking, decorations=None, parameters=None
):
    """
    Apply method for Petri net visualization (it calls the
    graphviz_visualization method)

    Parameters
    -----------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    decorations
        Decorations for elements in the Petri net
    parameters
        Algorithm parameters

    Returns
    -----------
    viz
        Graph object
    """
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    debug = exec_utils.get_param_value(Parameters.DEBUG, parameters, False)
    set_rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, None
    )
    font_size = exec_utils.get_param_value(
        Parameters.FONT_SIZE, parameters, "12"
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Petri Net"
    )

    if decorations is None:
        decorations = exec_utils.get_param_value(
            Parameters.DECORATIONS, parameters, None
        )

    return graphviz_visualization(
        net,
        image_format=image_format,
        initial_marking=initial_marking,
        final_marking=final_marking,
        decorations=decorations,
        debug=debug,
        set_rankdir=set_rankdir,
        font_size=font_size,
        bgcolor=bgcolor,
        enable_graph_title=enable_graph_title,
        graph_title=graph_title,
    )


def graphviz_visualization(
    net,
    image_format="png",
    initial_marking=None,
    final_marking=None,
    decorations=None,
    debug=False,
    set_rankdir=None,
    font_size="12",
    bgcolor=constants.DEFAULT_BGCOLOR,
    enable_graph_title: bool = constants.DEFAULT_ENABLE_GRAPH_TITLES,
    graph_title: str = "Petri Net",
):
    """
    Provides visualization for the petrinet

    Parameters
    ----------
    net: :class:`pm4py.entities.petri.petrinet.PetriNet`
        Petri net
    image_format
        Format that should be associated to the image
    initial_marking
        Initial marking of the Petri net
    final_marking
        Final marking of the Petri net
    decorations
        Decorations of the Petri net (says how element must be presented)
    debug
        Enables debug mode
    set_rankdir
        Sets the rankdir to LR (horizontal layout)
    enable_graph_title
        Enables the visualization of a graph's title
    graph_title
        Graph title to display (if enable_graph_title)

    Returns
    -------
    viz :
        Returns a graph object
    """
    if initial_marking is None:
        initial_marking = Marking()
    if final_marking is None:
        final_marking = Marking()
    if decorations is None:
        decorations = {}

    font_size = str(font_size)

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    viz = Digraph(
        net.name,
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor},
    )
    if set_rankdir:
        viz.graph_attr["rankdir"] = set_rankdir
    else:
        viz.graph_attr["rankdir"] = "LR"

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="'
            + str(2 * int(font_size))
            + '">'
            + graph_title
            + "</FONT>>",
            labelloc="top",
        )

    transitions = list(net.transitions)
    places = list(net.places)
    arcs = list(net.arcs)

    transitions, places, arcs = sort_petri_net(
        transitions, places, arcs, initial_marking, final_marking
    )

    viz.attr("node", shape="box")
    for t in transitions:
        label = (
            decorations[t]["label"]
            if t in decorations and "label" in decorations[t]
            else ""
        )
        fillcolor = (
            decorations[t]["color"]
            if t in decorations and "color" in decorations[t]
            else None
        )
        textcolor = "black"

        if t.label is not None and not label:
            label = t.label
        if debug:
            label = t.name
        label = str(label)

        if fillcolor is None:
            if t.label is None:
                fillcolor = "black"
                if label:
                    textcolor = "white"
            else:
                fillcolor = bgcolor

        viz.node(
            str(id(t)),
            label,
            style="filled",
            fillcolor=fillcolor,
            border="1",
            fontsize=font_size,
            fontcolor=textcolor,
        )

        if petri_properties.TRANS_GUARD in t.properties:
            guard = t.properties[petri_properties.TRANS_GUARD]
            viz.node(str(id(t)) + "guard", style="dotted", label=guard)
            viz.edge(
                str(id(t)) + "guard",
                str(id(t)),
                arrowhead="none",
                style="dotted",
            )

    for p in places:
        label = (
            decorations[p]["label"]
            if p in decorations and "label" in decorations[p]
            else ""
        )
        fillcolor = (
            decorations[p]["color"]
            if p in decorations and "color" in decorations[p]
            else bgcolor
        )

        label = str(label)
        if p in initial_marking:
            if initial_marking[p] == 1:
                viz.node(
                    str(id(p)),
                    "<&#9679;>",
                    fontsize="34",
                    fixedsize="true",
                    shape="circle",
                    width="0.75",
                    style="filled",
                    fillcolor=fillcolor,
                )
            else:
                marking_label = str(initial_marking[p])
                if len(marking_label) >= 3:
                    viz.node(
                        str(id(p)),
                        marking_label,
                        fontsize="34",
                        shape="ellipse",
                        style="filled",
                        fillcolor=fillcolor,
                    )
                else:
                    viz.node(
                        str(id(p)),
                        marking_label,
                        fontsize="34",
                        fixedsize="true",
                        shape="circle",
                        width="0.75",
                        style="filled",
                        fillcolor=fillcolor,
                    )
        elif p in final_marking:
            # <&#9632;>
            viz.node(
                str(id(p)),
                "<&#9632;>",
                fontsize="32",
                shape="doublecircle",
                fixedsize="true",
                width="0.75",
                style="filled",
                fillcolor=fillcolor,
            )
        else:
            if debug:
                viz.node(
                    str(id(p)),
                    str(p.name),
                    fontsize=font_size,
                    shape="ellipse",
                )
            else:
                if p in decorations and "label" in decorations[p]:
                    viz.node(
                        str(id(p)),
                        label,
                        style="filled",
                        fillcolor=fillcolor,
                        fontsize=font_size,
                        shape="ellipse",
                    )
                else:
                    viz.node(
                        str(id(p)),
                        label,
                        shape="circle",
                        fixedsize="true",
                        width="0.75",
                        style="filled",
                        fillcolor=fillcolor,
                    )

    # check if there is an arc with weight different than 1.
    # in that case, all the arcs in the visualization should have the arc
    # weight visible
    arc_weight_visible = False
    for arc in arcs:
        if arc.weight != 1:
            arc_weight_visible = True
            break

    for a in arcs:
        penwidth = (
            decorations[a]["penwidth"]
            if a in decorations and "penwidth" in decorations[a]
            else None
        )
        label = (
            decorations[a]["label"]
            if a in decorations and "label" in decorations[a]
            else ""
        )
        color = (
            decorations[a]["color"]
            if a in decorations and "color" in decorations[a]
            else None
        )

        if not label and arc_weight_visible:
            label = a.weight

        label = str(label)
        arrowhead = "normal"

        if petri_properties.ARCTYPE in a.properties:
            if (
                a.properties[petri_properties.ARCTYPE]
                == petri_properties.RESET_ARC
            ):
                arrowhead = "vee"
            elif (
                a.properties[petri_properties.ARCTYPE]
                == petri_properties.INHIBITOR_ARC
            ):
                arrowhead = "dot"

        viz.edge(
            str(id(a.source)),
            str(id(a.target)),
            label=label,
            penwidth=penwidth,
            color=color,
            fontsize=font_size,
            arrowhead=arrowhead,
            fontcolor=color,
        )

    viz.attr(overlap="false")

    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, Tuple

import pandas as pd

from pm4py.objects.ocel.obj import OCEL


def get_dataframes_from_ocel(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    if parameters is None:
        parameters = {}

    events = ocel.events.copy()
    for col in events.columns:
        if str(events[col].dtype) == "object":
            events[col] = events[col].astype("string")
        elif "date" in str(events[col].dtype):
            events[col] = events[col].dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    objects = ocel.objects.copy()
    for col in objects.columns:
        if str(objects[col].dtype) == "object":
            objects[col] = objects[col].astype("string")
        elif "date" in str(objects[col].dtype):
            objects[col] = objects[col].dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    return events, objects



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import xes_constants
from pm4py.util import constants
from pm4py.algo.discovery.dfg import algorithm as dfg_discovery
from pm4py.algo.discovery.causal import algorithm as causal_discovery
from pm4py.statistics.start_activities.log import get as get_start_activities
from pm4py.statistics.end_activities.log import get as get_end_activities
from pm4py.objects.conversion.log import converter
from pm4py.util import exec_utils
from enum import Enum
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog


class Outputs(Enum):
    DFG = "dfg"
    SEQUENCE = "sequence"
    PARALLEL = "parallel"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES = "activities"
    SKIPPABLE = "skippable"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_TRACE_LENGTH = "min_trace_length"
    TRACE = "trace"


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def apply(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, Any]:
    """
    Discovers a footprint object from an event log
    (the footprints of the event log are returned)

    Parameters
    --------------
    log
        Log
    parameters
        Parameters of the algorithm:
            - Parameters.ACTIVITY_KEY

    Returns
    --------------
    footprints_obj
        Footprints object
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    log = converter.apply(
        log, variant=converter.TO_EVENT_LOG, parameters=parameters
    )

    dfg = dfg_discovery.apply(log, parameters=parameters)
    parallel = {(x, y) for (x, y) in dfg if (y, x) in dfg}
    sequence = set(
        causal_discovery.apply(dfg, causal_discovery.Variants.CAUSAL_ALPHA)
    )

    start_activities = set(
        get_start_activities.get_start_activities(log, parameters=parameters)
    )
    end_activities = set(
        get_end_activities.get_end_activities(log, parameters=parameters)
    )
    activities = set(y[activity_key] for x in log for y in x)

    return {
        Outputs.DFG.value: dfg,
        Outputs.SEQUENCE.value: sequence,
        Outputs.PARALLEL.value: parallel,
        Outputs.START_ACTIVITIES.value: start_activities,
        Outputs.END_ACTIVITIES.value: end_activities,
        Outputs.ACTIVITIES.value: activities,
        Outputs.MIN_TRACE_LENGTH.value: (
            min(len(x) for x in log) if len(log) > 0 else 0
        ),
    }



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from scipy.spatial.distance import pdist
from pm4py.algo.discovery.dfg import algorithm as dfg_algorithm
from pm4py.statistics.attributes.log import get as attributes_filter
import pandas as pd
from pm4py.util import pandas_utils
from pm4py.algo.clustering.trace_attribute_driven.variants import act_dist_calc


def dfg_dist_calc_act(log1, log2):
    act1 = attributes_filter.get_attribute_values(log1, "concept:name")
    act2 = attributes_filter.get_attribute_values(log2, "concept:name")
    df1_act = act_dist_calc.occu_var_act(act1)
    df2_act = act_dist_calc.occu_var_act(act2)
    df_act = pandas_utils.merge(
        df1_act, df2_act, how="outer", on="var"
    ).fillna(0)
    dist_act = pdist(
        np.array([df_act["freq_x"].values, df_act["freq_y"].values]), "cosine"
    )[0]
    return dist_act


def dfg_dist_calc_suc(log1, log2):
    dfg1 = dfg_algorithm.apply(log1)
    dfg2 = dfg_algorithm.apply(log2)
    df1_dfg = act_dist_calc.occu_var_act(dfg1)
    df2_dfg = act_dist_calc.occu_var_act(dfg2)
    df_dfg = pandas_utils.merge(
        df1_dfg, df2_dfg, how="outer", on="var"
    ).fillna(0)
    dist_dfg = pdist(
        np.array([df_dfg["freq_x"].values, df_dfg["freq_y"].values]), "cosine"
    )[0]
    return dist_dfg


def dfg_dist_calc(log1, log2):
    act1 = attributes_filter.get_attribute_values(log1, "concept:name")
    act2 = attributes_filter.get_attribute_values(log2, "concept:name")
    dfg1 = dfg_algorithm.apply(log1)
    dfg2 = dfg_algorithm.apply(log2)
    df1_act = act_dist_calc.occu_var_act(act1)
    df2_act = act_dist_calc.occu_var_act(act2)
    df1_dfg = act_dist_calc.occu_var_act(dfg1)
    df2_dfg = act_dist_calc.occu_var_act(dfg2)
    df_act = pandas_utils.merge(
        df1_act, df2_act, how="outer", on="var"
    ).fillna(0)
    df_dfg = pandas_utils.merge(
        df1_dfg, df2_dfg, how="outer", on="var"
    ).fillna(0)
    dist_act = pdist(
        np.array([df_act["freq_x"].values, df_act["freq_y"].values]), "cosine"
    )[0]
    dist_dfg = pdist(
        np.array([df_dfg["freq_x"].values, df_dfg["freq_y"].values]), "cosine"
    )[0]
    if np.isnan(dist_dfg):
        dist_dfg = 1
    return dist_act, dist_dfg


def dfg_dist_calc_minkowski(log1, log2, alpha):
    act1 = attributes_filter.get_attribute_values(log1, "concept:name")
    act2 = attributes_filter.get_attribute_values(log2, "concept:name")
    dfg1 = dfg_algorithm.apply(log1)
    dfg2 = dfg_algorithm.apply(log2)
    df1_act = act_dist_calc.occu_var_act(act1)
    df2_act = act_dist_calc.occu_var_act(act2)
    df1_dfg = act_dist_calc.occu_var_act(dfg1)
    df2_dfg = act_dist_calc.occu_var_act(dfg2)
    df_act = pandas_utils.merge(
        df1_act, df2_act, how="outer", on="var"
    ).fillna(0)
    df_dfg = pandas_utils.merge(
        df1_dfg, df2_dfg, how="outer", on="var"
    ).fillna(0)
    dist_act = pdist(
        np.array(
            [
                df_act["freq_x"].values / np.sum(df_act["freq_x"].values),
                df_act["freq_y"].values / np.sum(df_act["freq_y"].values),
            ]
        ),
        "minkowski",
        p=2.0,
    )[0]
    dist_dfg = pdist(
        np.array(
            [
                df_dfg["freq_x"].values / np.sum(df_dfg["freq_x"].values),
                df_dfg["freq_y"].values / np.sum(df_dfg["freq_y"].values),
            ]
        ),
        "minkowski",
        p=2.0,
    )[0]
    dist = dist_act * alpha + dist_dfg * (1 - alpha)
    return dist



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any


class Parameters(Enum):
    INCLUDE_TIMESTAMPS = "include_timestamps"


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> str:
    """
    Provides a textual description of the provided object-centric event log

    Parameters
    --------------
    ocel
        Object-centric event log
    parameters
        Possible parameters of the algorithm, including:
        - Parameters.INCLUDE_TIMESTAMPS => include the timestamps (or not) in the representation

    Returns
    -------------
    ocel_stri
        Textual representation of the object-centric event log
    """
    if parameters is None:
        parameters = {}

    include_timestamps = exec_utils.get_param_value(
        Parameters.INCLUDE_TIMESTAMPS, parameters, True
    )

    object_ots = ocel.objects[
        [ocel.object_id_column, ocel.object_type_column]
    ].to_dict("records")
    object_ots = {
        x[ocel.object_id_column]: x[ocel.object_type_column]
        for x in object_ots
    }
    events = ocel.events.sort_values(
        [ocel.event_timestamp, ocel.event_activity, ocel.event_id_column]
    ).to_dict("records")
    objects = ocel.objects.sort_values(ocel.object_id_column).to_dict(
        "records"
    )
    relations = ocel.relations.sort_values(
        [
            ocel.event_timestamp,
            ocel.event_activity,
            ocel.object_id_column,
            ocel.event_id_column,
        ]
    )
    tdf = relations.groupby(ocel.object_id_column)[ocel.event_timestamp]
    objects_start = tdf.first().to_dict()
    objects_end = tdf.last().to_dict()
    objects_lifecycle = {
        x: objects_end[x].timestamp() - objects_start[x].timestamp()
        for x in objects_start
    }

    relations = (
        relations.groupby(ocel.event_id_column)[ocel.object_id_column]
        .agg(list)
        .to_dict()
    )

    ret = ["\n\nevents:\n"]

    for ev in events:
        stru = (
            ev[ocel.event_activity]
            + " ( related objects: "
            + ", ".join(relations[ev[ocel.event_id_column]])
            + " ) "
        )
        if include_timestamps:
            stru = stru + " timestamp: " + str(ev[ocel.event_timestamp])
        ret.append(stru)

    ret.append("\nobjects:\n")

    for obj in objects:
        obj_id = obj[ocel.object_id_column]
        stru = obj_id + " object type: " + object_ots[obj_id]
        if include_timestamps:
            stru = (
                stru
                + " ( lifecycle start: "
                + str(objects_start[obj_id])
                + " ; lifecycle end: "
                + str(objects_end[obj_id])
                + " ; lifecycle duration: "
                + str(objects_lifecycle[obj_id])
                + " )"
            )
        ret.append(stru)

    ret = "\n".join(ret)

    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.dfg.variants import native, performance
from pm4py.statistics.attributes.log import get as attr_get
from pm4py.util import xes_constants as xes
from pm4py.visualization.petri_net.common import visualize
from pm4py.visualization.petri_net.util.vis_trans_shortest_paths import (
    get_decorations_from_dfg_spaths_acticount,
)
from pm4py.visualization.petri_net.util.vis_trans_shortest_paths import (
    get_shortest_paths,
)
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ACTIVITY_KEY,
    PARAMETER_CONSTANT_TIMESTAMP_KEY,
)
from pm4py.objects.petri_net.obj import PetriNet, Marking
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    DEBUG = "debug"
    RANKDIR = "set_rankdir"
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    AGGREGATION_MEASURE = "aggregationMeasure"
    FONT_SIZE = "font_size"


def get_decorated_net(
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
    variant: str = "frequency",
) -> graphviz.Digraph:
    """
    Get a decorated net according to the specified variant (decorate Petri net based on DFG)

    Parameters
    ------------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    log
        Log to use to decorate the Petri net
    parameters
        Algorithm parameters
    variant
        Specify if the decoration should take into account the frequency or the performance

    Returns
    ------------
    gviz
        GraphViz object
    """
    if parameters is None:
        parameters = {}

    aggregation_measure = exec_utils.get_param_value(
        Parameters.AGGREGATION_MEASURE,
        parameters,
        "sum" if "frequency" in variant else "mean",
    )

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )

    # we find the DFG
    if variant == "performance":
        dfg = performance.performance(log, parameters=parameters)
    else:
        dfg = native.native(log, parameters=parameters)
    # we find shortest paths
    spaths = get_shortest_paths(net)
    # we find the number of activities occurrences in the log
    activities_count = attr_get.get_attribute_values(
        log, activity_key, parameters=parameters
    )
    aggregated_statistics = get_decorations_from_dfg_spaths_acticount(
        net,
        dfg,
        spaths,
        activities_count,
        variant=variant,
        aggregation_measure=aggregation_measure,
    )

    return visualize.apply(
        net,
        initial_marking,
        final_marking,
        parameters=parameters,
        decorations=aggregated_statistics,
    )


def apply(
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    log: EventLog = None,
    aggregated_statistics=None,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> graphviz.Digraph:
    """
    Apply frequency decoration through greedy algorithm (decorate Petri net based on DFG)

    Parameters
    ------------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    log
        Log to use to decorate the Petri net
    aggregated_statistics
        Dictionary containing the frequency statistics
    parameters
        Algorithm parameters

    Returns
    ------------
    gviz
        GraphViz object
    """
    del aggregated_statistics
    return get_decorated_net(
        net,
        initial_marking,
        final_marking,
        log,
        parameters=parameters,
        variant="frequency",
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import random
from typing import Counter, Generic, TypeVar

from pm4py.objects.petri_net.semantics import PetriNetSemantics
from pm4py.objects.petri_net.stochastic.obj import StochasticPetriNet

N = TypeVar("N", bound=StochasticPetriNet)
T = TypeVar("T", bound=StochasticPetriNet.Transition)
P = TypeVar("P", bound=StochasticPetriNet.Place)


class StochasticPetriNetSemantics(PetriNetSemantics[N], Generic[N]):

    @classmethod
    def sample_enabled_transition(
        cls, pn: N, marking: Counter[P], seed: int = None
    ) -> T:
        """
        Randomly samples a transition from all enabled transitions

        Parameters
        ----------
        :param pn: Petri net
        :param marking: marking to use

        Returns
        -------
        :return: a transition sampled from the enabled transitions
        """
        if seed is not None:
            random.seed(seed)
        enabled = list(
            filter(
                lambda t: cls.is_enabled(pn, t, marking),
                [t for t in pn.transitions],
            )
        )
        weights = list(
            map(
                lambda t: cls.probability_of_transition(pn, t, marking),
                enabled,
            )
        )
        return random.choices(enabled, weights)[0]

    @classmethod
    def probability_of_transition(
        cls, pn: N, transition: T, marking: Counter[P]
    ) -> float:
        """
        Compute the probability of firing a transition in the net and marking.

        Args:
            pn (N): Stochastic net
            transition (T): transition to fire
            marking (Counter[P]): marking to use

        Returns:
            float: _description_
        """
        if transition not in pn.transitions or not cls.is_enabled(
            pn, transition, marking
        ):
            return 0.0
        return transition.weight / sum(
            list(
                map(
                    lambda t: t.weight,
                    list(
                        filter(
                            lambda t: cls.is_enabled(pn, t, marking),
                            [t for t in pn.transitions],
                        )
                    ),
                )
            )
        )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.transition_system.variants import view_based
from pm4py.util import exec_utils
from enum import Enum
from pm4py.objects.transition_system.obj import TransitionSystem
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd


class Variants(Enum):
    VIEW_BASED = view_based


VERSIONS = {Variants.VIEW_BASED}
VIEW_BASED = Variants.VIEW_BASED
DEFAULT_VARIANT = Variants.VIEW_BASED


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> TransitionSystem:
    """
    Find transition system given log

    Parameters
    -----------
    log
        Log
    parameters
        Possible parameters of the algorithm, including:
            Parameters.PARAM_KEY_VIEW
            Parameters.PARAM_KEY_WINDOW
            Parameters.PARAM_KEY_DIRECTION
    variant
        Variant of the algorithm to use, including:
            Variants.VIEW_BASED

    Returns
    ----------
    ts
        Transition system
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(log, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.powl.inductive.variants.im_dynamic_clustering_frequencies import (
    POWLInductiveMinerDynamicClusteringFrequency, )
from pm4py.algo.discovery.powl.inductive.variants.im_tree import IMBasePOWL
from pm4py.algo.discovery.powl.inductive.variants.im_brute_force import (
    POWLInductiveMinerBruteForce,
)
from pm4py.algo.discovery.powl.inductive.variants.im_maximal import (
    POWLInductiveMinerMaximalOrder,
)
from pm4py.algo.discovery.powl.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant, )

from pm4py import util
from pm4py.algo.discovery.inductive.algorithm import Parameters
from pm4py.objects.powl.obj import POWL

from pm4py.util import xes_constants as xes_util
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL

from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, Type
from pm4py.objects.log.obj import EventLog
import pandas as pd


def get_variant(variant: POWLDiscoveryVariant) -> Type[IMBasePOWL]:
    if variant == POWLDiscoveryVariant.TREE:
        return IMBasePOWL
    elif variant == POWLDiscoveryVariant.BRUTE_FORCE:
        return POWLInductiveMinerBruteForce
    elif variant == POWLDiscoveryVariant.MAXIMAL:
        return POWLInductiveMinerMaximalOrder
    elif variant == POWLDiscoveryVariant.DYNAMIC_CLUSTERING:
        return POWLInductiveMinerDynamicClusteringFrequency
    else:
        raise Exception("Invalid Variant!")


def apply(
    obj: Union[EventLog, pd.DataFrame, UVCL],
    parameters: Optional[Dict[Any, Any]] = None,
    variant=POWLDiscoveryVariant.MAXIMAL,
) -> POWL:
    if parameters is None:
        parameters = {}
    ack = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_util.DEFAULT_NAME_KEY
    )
    tk = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes_util.DEFAULT_TIMESTAMP_KEY
    )
    cidk = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, util.constants.CASE_CONCEPT_NAME
    )
    if type(obj) in [EventLog, pd.DataFrame]:
        uvcl = comut.get_variants(
            comut.project_univariate(
                obj, key=ack, df_glue=cidk, df_sorting_criterion_key=tk
            )
        )
    else:
        uvcl = obj

    algorithm = get_variant(variant)
    im = algorithm(parameters)
    res = im.apply(IMDataStructureUVCL(uvcl), parameters)
    res = res.simplify()

    return res



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.conformance.tokenreplay import algorithm as token_replay
from pm4py.statistics.variants.log import get as variants_get
from pm4py.visualization.petri_net.common import visualize
from pm4py.visualization.petri_net.util import performance_map
from pm4py.util import exec_utils, xes_constants
from enum import Enum
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ACTIVITY_KEY,
    PARAMETER_CONSTANT_TIMESTAMP_KEY,
)
from pm4py.objects.petri_net.obj import PetriNet, Marking
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    DEBUG = "debug"
    RANKDIR = "set_rankdir"
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    AGGREGATION_MEASURE = "aggregationMeasure"
    FONT_SIZE = "font_size"


def get_decorations(
    log,
    net,
    initial_marking,
    final_marking,
    parameters=None,
    measure="frequency",
    ht_perf_method="last",
):
    """
    Calculate decorations in order to annotate the Petri net

    Parameters
    -----------
    log
        Trace log
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters associated to the algorithm
    measure
        Measure to represent on the process model (frequency/performance)
    ht_perf_method
        Method to use in order to annotate hidden transitions (performance value could be put on the last possible
        point (last) or in the first possible point (first)
    
    Returns
    ------------
    decorations
        Decorations to put on the process model
    """
    if parameters is None:
        parameters = {}

    aggregation_measure = exec_utils.get_param_value(
        Parameters.AGGREGATION_MEASURE, parameters, None
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    variants_idx = variants_get.get_variants_from_log_trace_idx(
        log, parameters=parameters
    )
    variants = variants_get.convert_variants_trace_idx_to_trace_obj(
        log, variants_idx
    )

    parameters_tr = {
        token_replay.Variants.TOKEN_REPLAY.value.Parameters.ACTIVITY_KEY: activity_key,
        token_replay.Variants.TOKEN_REPLAY.value.Parameters.VARIANTS: variants,
    }

    # do the replay
    aligned_traces = token_replay.apply(
        log, net, initial_marking, final_marking, parameters=parameters_tr
    )

    # apply petri_reduction technique in order to simplify the Petri net
    # net = reduction.apply(net, parameters={"aligned_traces": aligned_traces})

    element_statistics = performance_map.single_element_statistics(
        log,
        net,
        initial_marking,
        aligned_traces,
        variants_idx,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        ht_perf_method=ht_perf_method,
        parameters=parameters,
    )

    aggregated_statistics = performance_map.aggregate_statistics(
        element_statistics,
        measure=measure,
        aggregation_measure=aggregation_measure,
    )

    return aggregated_statistics


def apply(
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    log: EventLog = None,
    aggregated_statistics=None,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> graphviz.Digraph:
    """
    Apply method for Petri net visualization (it calls the graphviz_visualization
    method) adding frequency representation obtained by token replay

    Parameters
    -----------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    log
        (Optional) log
    aggregated_statistics
        Dictionary containing the frequency statistics
    parameters
        Algorithm parameters (including the activity key used during the replay, and the timestamp key)

    Returns
    -----------
    viz
        Graph object
    """
    if aggregated_statistics is None:
        if log is not None:
            aggregated_statistics = get_decorations(
                log,
                net,
                initial_marking,
                final_marking,
                parameters=parameters,
                measure="frequency",
            )
    return visualize.apply(
        net,
        initial_marking,
        final_marking,
        parameters=parameters,
        decorations=aggregated_statistics,
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.util import exec_utils, nx_utils
from pm4py.objects.conversion.log.variants import to_event_stream
from copy import copy


class Parameters(Enum):
    INCLUDE_DF = "include_df"
    INCLUDE_OBJECT_CHANGES = "include_object_changes"


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Converts an OCEL to a NetworkX DiGraph object.
    The nodes are the events and objects of the OCEL.
    The edges are of two different types:
    - relation edges, connecting an event to its related objects
    - directly-follows edges, connecting an event to a following event (in the lifecycle of one of the related objects)

    Parameters
    ---------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.INCLUDE_DF => include the directly-follows relationship

    Returns
    ---------------
    G
        NetworkX DiGraph
    """
    if parameters is None:
        parameters = {}

    include_df = exec_utils.get_param_value(
        Parameters.INCLUDE_DF, parameters, True
    )
    include_object_changes = exec_utils.get_param_value(
        Parameters.INCLUDE_OBJECT_CHANGES, parameters, True
    )

    G = nx_utils.DiGraph()

    stream = ocel.events.to_dict("records")
    stream = to_event_stream.__postprocess_stream(stream)
    for ev in stream:
        ev["type"] = "EVENT"
        G.add_node(ev[ocel.event_id_column], attr=ev)

    stream = ocel.objects.to_dict("records")
    stream = to_event_stream.__postprocess_stream(stream)
    for obj in stream:
        obj["type"] = "OBJECT"
        G.add_node(obj[ocel.object_id_column], attr=obj)

    rel_cols = {ocel.event_id_column, ocel.object_id_column, ocel.qualifier}

    relations = ocel.relations[list(rel_cols)]
    stream = relations.to_dict("records")
    for rel in stream:
        qualifier = rel[ocel.qualifier]
        if qualifier is None:
            qualifier = ""
        G.add_edge(
            rel[ocel.event_id_column],
            rel[ocel.object_id_column],
            attr={"type": "E2O", "qualifier": qualifier},
        )

    obj_relations = ocel.o2o[
        [ocel.object_id_column, ocel.object_id_column + "_2", ocel.qualifier]
    ]
    stream = obj_relations.to_dict("records")
    for rel in stream:
        qualifier = rel[ocel.qualifier]
        if qualifier is None:
            qualifier = ""
        G.add_edge(
            rel[ocel.object_id_column],
            rel[ocel.object_id_column + "_2"],
            attr={"type": "O2O", "qualifier": qualifier},
        )

    if include_df:
        lifecycle = (
            relations.groupby(ocel.object_id_column)
            .agg(list)
            .to_dict()[ocel.event_id_column]
        )
        for obj in lifecycle:
            lif = lifecycle[obj]
            for i in range(len(lif) - 1):
                G.add_edge(
                    lif[i], lif[i + 1], attr={"type": "DF", "object": obj}
                )

    if include_object_changes:
        object_changes = ocel.object_changes.to_dict("records")
        for i in range(len(object_changes)):
            change_id = "@@change##%d" % i
            change_dict = copy(object_changes[i])
            change_dict["type"] = "CHANGE"
            G.add_node(change_id, attr=change_dict)
            G.add_edge(
                change_id,
                object_changes[i][ocel.object_id_column],
                attr={"type": "CHANGE"},
            )

    return G



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import xes_constants as xes
from enum import Enum
from pm4py.util import constants, exec_utils
from pm4py.util import variants_util

from typing import Optional, Dict, Any, Union
from pm4py.objects.org.sna.obj import SNA
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    BETA = "beta"


BETA = Parameters.BETA


def apply(
    log: pd.DataFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> SNA:
    """
    Calculates the HW metric

    Parameters
    ------------
    log
        Log
    parameters
        Possible parameters of the algorithm:
            Paramters.BETA -> beta value as described in the Wil SNA paper

    Returns
    -----------
    tuple
        Tuple containing the metric matrix and the resources list. Moreover, last boolean indicates that the metric is
        directed.
    """
    if parameters is None:
        parameters = {}

    import numpy
    from pm4py.statistics.traces.generic.pandas import case_statistics

    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, xes.DEFAULT_RESOURCE_KEY
    )
    beta = exec_utils.get_param_value(Parameters.BETA, parameters, 0)
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    parameters_variants = {
        case_statistics.Parameters.ACTIVITY_KEY: resource_key,
        case_statistics.Parameters.ATTRIBUTE_KEY: resource_key,
        case_statistics.Parameters.CASE_ID_KEY: case_id_key,
    }

    variants_occ = {
        x["variant"]: x[case_id_key]
        for x in case_statistics.get_variant_statistics(
            log, parameters=parameters_variants
        )
    }
    variants_resources = list(variants_occ.keys())
    resources = [
        variants_util.get_activities_from_variant(y)
        for y in variants_resources
    ]

    flat_list = sorted(
        list(set([item for sublist in resources for item in sublist]))
    )

    metric_matrix = numpy.zeros((len(flat_list), len(flat_list)))

    sum_i_to_j = {}
    dividend = 0

    for idx, rv in enumerate(resources):
        rvj = variants_resources[idx]
        for i in range(len(rv) - 1):
            res_i = flat_list.index(rv[i])
            if res_i not in sum_i_to_j:
                sum_i_to_j[res_i] = {}
            for j in range(i + 1, len(rv)):
                res_j = flat_list.index(rv[j])
                if res_j not in sum_i_to_j[res_i]:
                    sum_i_to_j[res_i][res_j] = 0
                if beta == 0:
                    sum_i_to_j[res_i][res_j] += variants_occ[rvj]
                    dividend += variants_occ[rvj]
                    break
                else:
                    sum_i_to_j[res_i][res_j] += variants_occ[rvj] * (
                        beta ** (j - i - 1)
                    )
                    dividend += variants_occ[rvj] * (beta ** (j - i - 1))

    for key1 in sum_i_to_j:
        for key2 in sum_i_to_j[key1]:
            metric_matrix[key1][key2] = sum_i_to_j[key1][key2] / dividend

    connections = {}
    for key1 in sum_i_to_j:
        for key2 in sum_i_to_j[key1]:
            connections[(flat_list[key1], flat_list[key2])] = (
                sum_i_to_j[key1][key2] / dividend
            )
            metric_matrix[key1][key2] = sum_i_to_j[key1][key2] / dividend

    return SNA(connections, True)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from copy import deepcopy


def apply(ocel: OCEL) -> OCEL:
    """
    Rename objects given their object type, lifecycle start/end timestamps, and lexicographic order,

    Parameters
    -----------------
    ocel
        Object-centric event log

    Returns
    ----------------
    renamed_ocel
        Object-centric event log with renaming
    """
    objects_start = (
        ocel.relations.groupby(ocel.object_id_column)[ocel.event_timestamp]
        .first()
        .to_dict()
    )
    objects_end = (
        ocel.relations.groupby(ocel.object_id_column)[ocel.event_timestamp]
        .last()
        .to_dict()
    )
    objects_ot0 = ocel.objects[
        [ocel.object_id_column, ocel.object_type_column]
    ].to_dict("records")
    objects_ot0 = [
        (x[ocel.object_id_column], x[ocel.object_type_column])
        for x in objects_ot0
    ]
    objects_ot1 = {}

    for el in objects_ot0:
        if not el[1] in objects_ot1:
            objects_ot1[el[1]] = []
        objects_ot1[el[1]].append(el[0])

    overall_objects = {}
    keys = sorted(list(objects_ot1))
    for ot in keys:
        objects = objects_ot1[ot]
        objects.sort(key=lambda x: (objects_start[x], objects_end[x], x))
        objects = {
            objects[i]: ot + "_" + str(i + 1) for i in range(len(objects))
        }
        overall_objects.update(objects)

    ocel = deepcopy(ocel)
    ocel.objects[ocel.object_id_column] = ocel.objects[
        ocel.object_id_column
    ].map(overall_objects)
    ocel.relations[ocel.object_id_column] = ocel.relations[
        ocel.object_id_column
    ].map(overall_objects)
    ocel.o2o[ocel.object_id_column] = ocel.o2o[ocel.object_id_column].map(
        overall_objects
    )
    ocel.o2o[ocel.object_id_column + "_2"] = ocel.o2o[
        ocel.object_id_column + "_2"
    ].map(overall_objects)
    ocel.object_changes[ocel.object_id_column] = ocel.object_changes[
        ocel.object_id_column
    ].map(overall_objects)

    return ocel
